/*
Module : PJNHash.h
Purpose: Defines the interface for some classes which encapsulate calculating Cryptographic hashes and HMACS using CryptoAPI: Next Generation (CNG)
Created: PJN / 23-04-2005
History: PJN / 18-05-2005 1. Fixed a compiler warning when compiled using Visual Studio .NET 2003. Thanks to Alexey Kuznetsov
                          for reporting this issue.
         PJN / 29-09-2005 1. Format method now allows you to specify an uppercase or lower case string for the hash. This is 
                          necessary since the CRAM-MD5 authentication mechanism requires a lowercase MD5 hash. Thanks to 
                          Jian Peng for reporting this issue.
         PJN / 16-01-2015 1. Updated copyright details
         PJN / 11-08-2016 1. Removed defunct comments about VC 6 in CPJNMD5 class.
                          2. Reworked the class to optionally compile without MFC. By default the class now uses STL 
                          classes and idioms but if you define CPJNMD5_MFC_EXTENSIONS the classes will revert back to the 
                          MFC behaviour.
         PJN / 24-06-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
         PJN / 22-08-2018 1. Fixed a number of compiler warnings when using VS 2018 15.8.1
         PJN / 07-04-2019 1. Updated copyright details.
                          2. Removed code which supports the now defunct CPJNMD5_MFC_EXTENSIONS codepath
         PJN / 14-04-2019 1. Replaced BOOL with bool throughout the code.
         PJN / 27-12-2020 1. Fixed a bug in CPJNMD5Hash::Format where the size of the string was not being set correctly.
                          2. Reimplemented CPJNMD5Hash::Format using CryptBinaryToString.
                          3. Reimplemented CPJNMD5 class using CryptoAPI: Next Generation (CNG) instead of the older MS CryptoAPI
                          4. Renamed CPJNMD5Hash class to CPJNHashResult.
                          5. Renamed CPJNMD5 class to CPJNHash.
                          6. Reworked the CPJNHash to support any hash algorithm supported by CNG.
                          7. Renamed PJNMD5.h module to PJNHash.h
         PJN / 03-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations
         PJN / 26-05-2023 1. Updated copyright details.
                          2. Updated module to indicate that it needs to be compiled using /std:c++17. Thanks to Martin Richter 
                          for reporting this issue.

Copyright (c) 2005 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


////////////////////////////// Macros / Defines ///////////////////////////////

#pragma once

#if _MSVC_LANG < 201703
#error CPJNHash requires a minimum C++ language standard of /std:c++17
#endif //#if _MSVC_LANG < 201703

#ifndef __PJNHASH_H__
#define __PJNHASH_H__


////////////////////////////// Includes ///////////////////////////////////////

#ifndef __BCRYPT_H__
#pragma message("To avoid this message, please put bcrypt.h in your pre compiled header (normally stdafx.h)")
#include <bcrypt.h>
#endif //#ifndef __BCRYPT_H__


////////////////////////////// Classes ////////////////////////////////////////

// A simple wrapper class which contains a hash and provides a method to format it as a string
class CPJNHashResult : public std::vector<BYTE>
{
public:
//typedefs
#ifdef _UNICODE
  using String = std::wstring;
#else
  using String = std::string;
#endif //#ifdef _UNICODE

//Methods
  String Format()
  {
#pragma warning(suppress: 26472)
    const auto dwSize{static_cast<DWORD>(size())};
    const auto pbyData{data()};
    DWORD cchString{0};
    if (!CryptBinaryToString(pbyData, dwSize, CRYPT_STRING_HEXRAW | CRYPT_STRING_NOCRLF, nullptr, &cchString))
      return {};
    String sRet;
    sRet.resize(cchString);
    if (!CryptBinaryToString(pbyData, dwSize, CRYPT_STRING_HEXRAW | CRYPT_STRING_NOCRLF, sRet.data(), &cchString))
      return {};
    sRet.resize(cchString);
    return sRet;
  }
};


// Class which calculates Hashes and HMACS
class CPJNHash
{
public:
//Constructors / Destructors
  CPJNHash(_In_ LPCWSTR pszAlgId) noexcept : m_hProv{nullptr},
                                             m_StatusBCryptOpen{0}
  {
    m_StatusBCryptOpen = BCryptOpenAlgorithmProvider(&m_hProv, pszAlgId, nullptr, 0);
  }

  CPJNHash(const CPJNHash&) = delete;
  CPJNHash(CPJNHash&&) = delete;

  virtual ~CPJNHash()
  {
    if (m_hProv)
      BCryptCloseAlgorithmProvider(m_hProv, 0);
  }

//Methods
  CPJNHash& operator=(const CPJNHash&) = delete;
  CPJNHash& operator=(CPJNHash&&) = delete;

  //Simple hash function which hashes a blob of data and returns the results encapsulated in the "hash" parameter
  NTSTATUS Hash(_In_reads_bytes_(dwDataSize) const BYTE* pbyData, _In_ DWORD dwDataSize, _Inout_ CPJNHashResult& hash)
  {
    if (!BCRYPT_SUCCESS(m_StatusBCryptOpen))
      return m_StatusBCryptOpen;

    //Call the Helper Hash function which returns a BCRYPT_HASH_HANDLE handle which contains the hash
    BCRYPT_HASH_HANDLE hHash{nullptr};
    NTSTATUS status{Hash(pbyData, dwDataSize, hHash)};
    if (!BCRYPT_SUCCESS(status))
    {
      if (hHash)
        BCryptDestroyHash(hHash);
      return status;
    }

    //Get the hash size
    DWORD dwHashSize{0};
    DWORD dwPropertySize{sizeof(dwHashSize)};
#pragma warning(suppress: 26490)
    status = BCryptGetProperty(hHash, BCRYPT_HASH_LENGTH, reinterpret_cast<PUCHAR>(&dwHashSize), sizeof(dwHashSize), &dwPropertySize, 0);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hHash);
      return status;
    }

    //Get the hash data
    hash.resize(dwHashSize);
#pragma warning(suppress: 26472)
    status = BCryptFinishHash(hHash, hash.data(), static_cast<ULONG>(hash.size()), 0);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hHash);
      return status;
    }

    //Free up the hash handle now that we are finished with it
    BCryptDestroyHash(hHash);

    return status;
  }

  //Function which returns a hash encoded as a BCRYPT_HASH_HANDLE. If hHash is
  //nullptr on entry then a new BCRYPT_HASH_HANDLE is created. If hHash is non-null
  //then this existing hash is used and the data to hash is included into
  //the current hash. The client is responsible for freeing the hash handle
  NTSTATUS Hash(_In_reads_bytes_(dwDataSize) const BYTE* pbyData, DWORD dwDataSize, _Inout_ BCRYPT_HASH_HANDLE& hHash) noexcept
  {
    //Create the hash object if required to
    if (hHash == nullptr)
    {
      const NTSTATUS status{BCryptCreateHash(m_hProv, &hHash, nullptr, 0, nullptr, 0, 0)};
      if (!BCRYPT_SUCCESS(status))
        return status;
    }

    //Hash in the data
#pragma warning(suppress: 26465 26492)
    return BCryptHashData(hHash, const_cast<PUCHAR>(pbyData), dwDataSize, 0);
  }

  //Function which calculates a keyed HMAC according to RFC 2104
  NTSTATUS HMAC(_In_reads_bytes_(dwDataSize) const BYTE* pbyData, _In_ DWORD dwDataSize, _In_reads_bytes_(dwKeySize) const BYTE* pbyKey, _In_ DWORD dwKeySize, _Inout_ CPJNHashResult& hash)
  {
    //if key is longer than 64 bytes then reset it to the hash of the key
    const BYTE* pbyLocalKey{nullptr};
    DWORD dwLocalKeySize{0};
    CPJNHashResult keyHash;
    if (dwKeySize > 64)
    {
      const NTSTATUS status{Hash(pbyKey, dwKeySize, keyHash)};
      if (!BCRYPT_SUCCESS(status))
        return status;
#pragma warning(suppress: 26472)
      dwLocalKeySize = static_cast<DWORD>(keyHash.size());
      pbyLocalKey = keyHash.data();
    }
    else
    {
      dwLocalKeySize = dwKeySize;
      pbyLocalKey = pbyKey;
    }

    //start out by storing key in pads
    std::vector<BYTE> k_ipad{64, std::allocator<BYTE>{}};
    std::vector<BYTE> k_opad{64, std::allocator<BYTE>{}};
    memcpy_s(k_ipad.data(), k_ipad.size(), pbyLocalKey, dwLocalKeySize);
    memcpy_s(k_opad.data(), k_opad.size(), pbyLocalKey, dwLocalKeySize);

    //XOR key with ipad value
    for (auto& element : k_ipad)
      element ^= 0x36;

    //XOR key with opad
    for (auto& element : k_opad)
      element ^= 0x5c;

    //perform inner hash
    BCRYPT_HASH_HANDLE hInnerHash{nullptr};
  #pragma warning(suppress: 6385)
#pragma warning(suppress: 26472)
    NTSTATUS status{Hash(k_ipad.data(), static_cast<DWORD>(k_ipad.size()), hInnerHash)};
    if (!BCRYPT_SUCCESS(status))
    {
      if (hInnerHash)
        BCryptDestroyHash(hInnerHash);
      return status;
    }
    status = Hash(pbyData, dwDataSize, hInnerHash);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hInnerHash);
      return status;
    }

    //Get the inner hash result
#pragma warning(suppress: 26472)
    DWORD dwHashSize{0};
    DWORD dwPropertySize{sizeof(dwHashSize)};
#pragma warning(suppress: 26490)
    status = BCryptGetProperty(hInnerHash, BCRYPT_HASH_LENGTH, reinterpret_cast<PUCHAR>(&dwHashSize), sizeof(dwHashSize), &dwPropertySize, 0);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hInnerHash);
      return status;
    }
    CPJNHashResult InnerHash;
    InnerHash.resize(dwHashSize);
#pragma warning(suppress: 26472)
    status = BCryptFinishHash(hInnerHash, InnerHash.data(), static_cast<ULONG>(InnerHash.size()), 0);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hInnerHash);
      return status;
    }

    //Free up the inner hash now that we are finished with it
    BCryptDestroyHash(hInnerHash);

    //Perform outer hash
    BCRYPT_HASH_HANDLE hOuterHash{nullptr};
#pragma warning(suppress: 26472)
    status = Hash(k_opad.data(), static_cast<DWORD>(k_opad.size()), hOuterHash);
    if (!BCRYPT_SUCCESS(status))
    {
      if (hOuterHash)
        BCryptDestroyHash(hOuterHash);
      return false;
    }
#pragma warning(suppress: 26472)
    status = Hash(InnerHash.data(), static_cast<DWORD>(InnerHash.size()), hOuterHash);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hOuterHash);
      return false;
    }

    //Finally get the hash result
#pragma warning(suppress: 26472)
    dwHashSize = 0;
    dwPropertySize = sizeof(dwHashSize);
#pragma warning(suppress: 26490)
    status = BCryptGetProperty(hOuterHash, BCRYPT_HASH_LENGTH, reinterpret_cast<PUCHAR>(&dwHashSize), sizeof(dwHashSize), &dwPropertySize, 0);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hOuterHash);
      return status;
    }
    hash.resize(dwHashSize);
#pragma warning(suppress: 26472)
    status = BCryptFinishHash(hOuterHash, hash.data(), static_cast<ULONG>(hash.size()), 0);
    if (!BCRYPT_SUCCESS(status))
    {
      BCryptDestroyHash(hOuterHash);
      return status;
    }

    BCryptDestroyHash(hOuterHash);
    return status;
  }

protected:
//Member variables
  BCRYPT_ALG_HANDLE m_hProv;
  NTSTATUS m_StatusBCryptOpen;
};

#endif //#ifndef __PJNHASH_H__
